@@ -1,34 +1,42 @@
-# $Id: crt.py 1029 2014-08-31 12:57:33Z mwall $
+# $Id: crt.py 1219 2015-01-14 16:15:27Z mwall $
 # Copyright 2013 Matthew Wall
+"""Emit loop data to file in Cumulus realtime format.
+   http://wiki.sandaysoft.com/a/Realtime.txt
 
-# Emit loop data to file in Cumulus realtime format.
-#   http://wiki.sandaysoft.com/a/Realtime.txt
-#
-# Put this file in bin/user/crt.py, then add this to your weewx.conf:
-#
-# [CumulusRealTime]
-#     date_separator = /
-#     filename = /path/to/realtime.txt
-#
-# [Engines]
-#     [[WxEngine]]
-#         process_services = ..., user.crt.CumulusRealTime
-#
-# Note that most of the code in this file is to calculate/lookup data that
-# are not directly provided by weewx in a LOOP packet.
-#
-# The cumulus 'specification' for realtime.txt is ambiguous:
-#
-#   pressure trend interval is not specified, we use 3 hours for pressure
-#   temperature trend interval is not specified, we use 3 hours for temperature
-#   wind avg speed/dir interval is not specified, we use 10 minutes
-#   time zone is not specified, we use utc everywhere
-#   pressure is not specified, we use 'barometer' not 'pressure' or 'altimeter'
-#
-#   FIXME: what if None?  we return NULL, but cumulus spec does not specify
-#   FIXME: is_sunny is ill-defined.  what radiation threshold indicates true?
-#   FIXME: ensure day/month/hour/minute/second is prefixed with leading zero
-#   FIXME: max_rad is not yet implemented
+Put this file in bin/user/crt.py, then add this to your weewx.conf:
+
+[CumulusRealTime]
+    unit_system = METRIC # options are US, METRIC, METRICWX
+    date_separator = /
+    filename = /path/to/realtime.txt
+
+[Engine]
+    [[Services]]
+        archive_services = ..., user.crt.CumulusRealTime
+
+If no unit_system is specified, the units will be those of the database.
+
+Note that most of the code in this file is to calculate/lookup data that
+are not directly provided by weewx in a LOOP packet.
+
+The cumulus 'specification' for realtime.txt is ambiguous:
+
+  pressure trend interval is not specified, we use 3 hours for pressure
+  temperature trend interval is not specified, we use 3 hours for temperature
+  wind avg speed/dir interval is not specified, we use 10 minutes
+  time zone is not specified, we use utc everywhere
+  pressure is not specified, we use 'barometer' not 'pressure' or 'altimeter'
+
+  how to handle None/NULL is not specified.  we return a NULL when the value
+  is None, but the cumulus specification does not indicate a correct
+  behavior.
+"""
+
+# FIXME: is_sunny is ill-defined.  what radiation threshold indicates true?
+# FIXME: ensure day/month/hour/minute/second is prefixed with leading zero
+# FIXME: max_rad is not yet implemented
+# FIXME: is rainYear past 365 days or since 1 jan?
+# FIXME: is rainMonth past 30 days or since 1st of month?
 
 import math
 import time
@@ -38,40 +46,42 @@ import weewx
 import weewx.wxformulas
 import weeutil.weeutil
 import weeutil.Sun
-from weewx.wxengine import StdService
+from weewx.engine import StdService
 
-VERSION = 'X'
+VERSION = "0.10"
+
+if weewx.__version__ < "3":
+    raise weewx.UnsupportedFeature("weewx 3 is required, found %s" %
+                                   weewx.__version__)
+
+def logmsg(level, msg):
+    syslog.syslog(level, 'crt: %s' % msg)
+
+def logdbg(msg):
+    logmsg(syslog.LOG_DEBUG, msg)
+
+def loginf(msg):
+    logmsg(syslog.LOG_INFO, msg)
+
+def logerr(msg):
+    logmsg(syslog.LOG_ERR, msg)
 
 COMPASS_POINTS = ['N','NNE','NE','ENE','E','ESE','SE','SSE',
                   'S','SSW','SW','WSW','W','WNW','NW','NNW','N']
 
 # map weewx unit names to cumulus unit names
-UNITS_WIND = {'mile_per_hour':      'mph',
-              'meter_per_second':   'm/s',
+UNITS_WIND = {'mile_per_hour': 'mph',
+              'meter_per_second': 'm/s',
               'kilometer_per_hour': 'km/h',
-              'knot':               'kts'}
+              'knot': 'kts'}
 UNITS_TEMP = {'degree_C': 'C',
               'degree_F': 'F'}
 UNITS_PRES = {'inHg': 'in',
               'mbar': 'mb',
-              'hPa':  'hPa'}
+              'hPa': 'hPa'}
 UNITS_RAIN = {'in': 'in',
               'mm': 'mm'}
 
-def getAltitudeFt(config_dict):
-    """Get the altitude, in meters, from the Station section of the dict."""
-    altitude_t = weeutil.weeutil.option_as_list(
-        config_dict['Station'].get('altitude', (None, None)))
-    altitude_vt = (float(altitude_t[0]), altitude_t[1], "group_altitude")
-    altitude_f = weewx.units.convert(altitude_vt, 'foot')[0]
-    return altitude_f
-
-def getLatitude(config_dict):
-    return float(config_dict['Station'].get('latitude', None))
-
-def getLongitude(config_dict):
-    return float(config_dict['Station'].get('longitude', None))
-
 def degreeToCompass(x):
     if x is None:
         return None
@@ -108,118 +118,103 @@ def knotToBeaufort(x):
         return 11 # violent storm
     return 12 # hurricane
 
-def calcAvgWindSpeed(archive, ts, interval=600):
+def calcAvgWindSpeed(dbm, ts, interval=600):
     sts = ts - interval
-    val = archive.getSql("SELECT AVG(windSpeed) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT AVG(windSpeed) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
-def calcAvgWindDir(archive, ts, interval=600):
+def calcAvgWindDir(dbm, ts, interval=600):
     sts = ts - interval
-    val = archive.getSql("SELECT AVG(windDir) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT AVG(windDir) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
-def calc10MinHighGust(archive, ts):
+def calc10MinHighGust(dbm, ts):
     sts = ts - 600
-    val = archive.getSql("SELECT MAX(windGust) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT MAX(windGust) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
-def calc10MinAvgWindDir(archive, ts):
+def calc10MinAvgWindDir(dbm, ts):
     sts = ts - 600
-    val = archive.getSql("SELECT AVG(windGust) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT AVG(windDir) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
 # calculate wind run since midnight
-def calcWindRun(archive, ts):
+def calcWindRun(dbm, ts):
     run = 0
     sod_ts = weeutil.weeutil.startOfDay(ts)
-    for row in archive.genSql("SELECT `interval`,windSpeed FROM archive "
-                              "WHERE dateTime>? AND dateTime<=?", 
-                              (sod_ts, ts)):
+    for row in dbm.genSql("SELECT `interval`,windSpeed FROM %s "
+                          "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                          (sod_ts, ts)):
         if row[1] is not None:
             run += row[1] * (row[0] / 60)
     return run
 
 # calculate trend over past n hours, default to 3 hour window
-def calcTrend(label, new_val, archive, ts, n=3):
+def calcTrend(label, new_val, dbm, ts, n=3):
     if new_val is None:
         return None
     lastts = ts - n * 3600
-    qstr = "SELECT %s FROM archive WHERE dateTime>? AND dateTime<=?" % label
-    old_val = archive.getSql(qstr, (lastts, ts))
-    if old_val is None:
+    old_val = dbm.getSql("SELECT %s FROM %s "
+                         "WHERE dateTime>? AND dateTime<=?" %
+                         (label, dbm.table_name), (lastts, ts))
+    if old_val is None or old_val[0] is None:
         return None
     return new_val - old_val[0]
 
-def calcRainHour(archive, ts):
+def calcRainHour(dbm, ts):
     sts = ts - 3600
-    val = archive.getSql("SELECT SUM(rain) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT SUM(rain) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
-def calcRainMonth(archive, ts):
+def calcRainMonth(dbm, ts):
     span = weeutil.weeutil.archiveMonthSpan(ts)
-    val = archive.getSql("SELECT SUM(rain) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
+    val = dbm.getSql("SELECT SUM(rain) FROM %s "
+                         "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
                          (span.start, ts))
-    if val is None:
-        return None
-    return val[0]
+    return val[0] if val is not None else None
 
-def calcRainYear(archive, ts):
+def calcRainYear(dbm, ts):
     span = weeutil.weeutil.archiveYearSpan(ts)
-    val = archive.getSql("SELECT SUM(rain) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (span.start, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT SUM(rain) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (span.start, ts))
+    return val[0] if val is not None else None
 
-def calcRainYesterday(archive, ts):
+def calcRainYesterday(dbm, ts):
     ts = weeutil.weeutil.startOfDay(ts)
     sts = ts - 3600*24;
-    val = archive.getSql("SELECT SUM(rain) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT SUM(rain) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
-def calcETToday(archive, ts):
+def calcETToday(dbm, ts):
     sts = weeutil.weeutil.startOfDay(ts)
-    val = archive.getSql("SELECT SUM(ET) FROM archive "
-                         "WHERE dateTime>? AND dateTime<=?",
-                         (sts, ts))
-    if val is None:
-        return None
-    return val[0]
+    val = dbm.getSql("SELECT SUM(ET) FROM %s "
+                     "WHERE dateTime>? AND dateTime<=?" % dbm.table_name,
+                     (sts, ts))
+    return val[0] if val is not None else None
 
-def calcMinMax(label, archive, ts, minmax='MAX'):
+def calcMinMax(label, dbm, ts, minmax='MAX'):
     sts = weeutil.weeutil.startOfDay(ts)
-    qstr = "SELECT %s(%s) FROM archive WHERE dateTime>=? AND dateTime<=?" % (minmax,label)
-    val = archive.getSql(qstr, (sts, ts))
+    val = dbm.getSql("SELECT %s(%s) FROM %s "
+                     "WHERE dateTime>=? AND dateTime<=?" %
+                     (minmax, label, dbm.table_name), (sts, ts))
     if val is None:
         return None, None
-    qstr = "SELECT dateTime FROM archive WHERE dateTime>=? AND dateTime<=? AND %s=?" % label
-    t = archive.getSql(qstr, (sts, ts, val[0]))
+    t = dbm.getSql("SELECT dateTime FROM %s "
+                   "WHERE dateTime>=? AND dateTime<=? AND %s=?" %
+                   (dbm.table_name, label), (sts, ts, val[0]))
     if t is None:
         return None, None
     tstr = time.strftime("%H:%M", time.gmtime(t[0]))
@@ -262,7 +257,7 @@ def calcCloudBase(outTemp_F, dewpoint_F, altitude_foot):
         return None
     return int((outTemp_F - dewpoint_F) / 4.4 * 1000) + altitude_foot
 
-# this depends on the weather station
+# this depends on the weather station.  many stations have more than one.
 #
 # Vantage
 #   packet['rxCheckPercent'] == 0
@@ -286,35 +281,65 @@ def calcCloudBase(outTemp_F, dewpoint_F, altitude_foot):
 # WS23xx
 #   packet['cn'] == 'lost'
 #
-def lostSensorContact():
-    # FIXME: implement lostSensorContact
+def lostSensorContact(packet):
+    if 'rxCheckPercent' in packet and packet['rxCheckPercent'] == 0:
+        return 1
+    if 'cn' in packet and packet['cn'] == 'lost':
+        return 1
+    if (('windspeed_state' in packet and packet['windspeed_state'] == 'no_link') or
+        ('rain_state' in packet and packet['rain_state'] == 'no_link') or
+        ('uv_state' in packet and packet['uv_state'] == 'no_link') or
+        ('h_1_state' in packet and packet['h_1_state'] == 'no_link') or
+        ('h_2_state' in packet and packet['h_2_state'] == 'no_link') or
+        ('h_3_state' in packet and packet['h_3_state'] == 'no_link') or
+        ('h_4_state' in packet and packet['h_4_state'] == 'no_link') or
+        ('h_5_state' in packet and packet['h_5_state'] == 'no_link')):
+        return 1
     return 0
 
+
 class CumulusRealTime(StdService):
 
     def __init__(self, engine, config_dict):
         super(CumulusRealTime, self).__init__(engine, config_dict)
-        try:
-            db = config_dict['StdArchive']['archive_database']
-            self.database_dict = config_dict['Databases'][db]
-        except KeyError, e:
-            syslog.syslog(syslog.LOG_DEBUG, 
-                          "crt: cannot determine database configuration: " % e)
-            return
-        self.altitude_ft = getAltitudeFt(config_dict)
-        self.latitude = getLatitude(config_dict)
-        self.longitude = getLongitude(config_dict)
+        loginf("service version is %s" % VERSION)
+        self.altitude_ft = weewx.units.convert(engine.stn_info.altitude_vt,
+                                               "foot")[0]
+        self.latitude = engine.stn_info.latitude_f
+        self.longitude = engine.stn_info.longitude_f
 
         d = config_dict.get('CumulusRealTime', {})
         self.filename = d.get('filename', '/var/tmp/realtime.txt')
         self.datesep = d.get('date_separator', '/')
-        self.bind(weewx.NEW_LOOP_PACKET, self.handle_new_loop)
+        self.binding = d.get('binding', 'loop')
+        us = None
+        us_label = d.get('unit_system', None)
+        if us_label is not None:
+            if us_label in weewx.units.unit_constants:
+                loginf("units will be displayed as %s" % us_label)
+                us = weewx.units.unit_constants[us_label]
+            else:
+                logerr("unknown unit_system %s" % us_label)
+        self.unit_system = us
+
+        if self.binding == 'loop':
+            self.bind(weewx.NEW_LOOP_PACKET, self.handle_new_loop)
+        else:
+            self.bind(weewx.NEW_ARCHIVE_RECORD, self.handle_new_archive)
+
+        loginf("binding is %s" % self.binding)
+        loginf("output goes to %s" % self.filename)
 
     def handle_new_loop(self, event):
+        self.handle_data(event.packet)
+
+    def handle_new_archive(self, event):
+        self.handle_data(event.record)
+
+    def handle_data(self, event_data):
         try:
-            data = {}
-            with weewx.archive.Archive.open(self.database_dict) as archive:
-                data = self.calculate(event.packet, archive)
+            dbm = self.engine.db_binder.get_manager('wx_binding')
+            data = self.calculate(event_data, dbm)
             self.write_data(data)
         except Exception, e:
             weeutil.weeutil.log_traceback('crt: **** ')
@@ -326,11 +351,15 @@ class CumulusRealTime(StdService):
 
     # calculate the data elements that cumulus needs that weewx does not
     # provide directly.
-    def calculate(self, packet, archive):
+    def calculate(self, packet, dbm):
         ts = packet.get('dateTime')
         pu = packet.get('usUnits')
+        if self.unit_system is not None and self.unit_system != pu:
+            packet = weewx.units.to_std_system(packet, self.unit_system)
+            pu = self.unit_system
+
         data = dict(packet)
-        data['windSpeed_avg'] = calcAvgWindSpeed(archive, ts)
+        data['windSpeed_avg'] = calcAvgWindSpeed(dbm, ts)
         data['windDir_compass'] = degreeToCompass(packet.get('windDir'))
         x = packet.get('windSpeed')
         ut = weewx.units.getStandardUnitType(pu, 'windSpeed')
@@ -339,40 +368,39 @@ class CumulusRealTime(StdService):
         data['windSpeed_beaufort'] = knotToBeaufort(v)
         data['units_wind'] = UNITS_WIND.get(ut[0])
         ut = weewx.units.getStandardUnitType(pu, 'outTemp')
+        logerr("ut: %s pu: %s" % (ut, pu))
         data['units_temperature'] = UNITS_TEMP.get(ut[0],ut[0])
         ut = weewx.units.getStandardUnitType(pu, 'barometer')
         data['units_pressure'] = UNITS_PRES.get(ut[0],ut[0])
         ut = weewx.units.getStandardUnitType(pu, 'rain')
         data['units_rain'] = UNITS_RAIN.get(ut[0],ut[0])
-        data['wind_run'] = calcWindRun(archive, ts)
+        data['wind_run'] = calcWindRun(dbm, ts)
         data['pressure_trend'] = calcTrend('barometer',
-                                           packet.get('barometer'),
-                                           archive, ts)
+                                           packet.get('barometer'), dbm, ts)
         data['temperature_trend'] = calcTrend('outTemp',
-                                              packet.get('outTemp'),
-                                              archive, ts)
-        data['rain_month'] = calcRainMonth(archive, ts)
-        data['rain_year'] = calcRainYear(archive, ts)
-        data['rain_yesterday'] = calcRainYesterday(archive, ts)
-        v,t = calcMinMax('outTemp', archive, ts, 'MAX')
+                                              packet.get('outTemp'), dbm, ts)
+        data['rain_month'] = calcRainMonth(dbm, ts)
+        data['rain_year'] = calcRainYear(dbm, ts)
+        data['rain_yesterday'] = calcRainYesterday(dbm, ts)
+        v,t = calcMinMax('outTemp', dbm, ts, 'MAX')
         data['outTemp_max'] = v
         data['outTemp_max_time'] = t
-        v,t = calcMinMax('outTemp', archive, ts, 'MIN')
+        v,t = calcMinMax('outTemp', dbm, ts, 'MIN')
         data['outTemp_min'] = v
         data['outTemp_min_time'] = t
-        v,t = calcMinMax('windSpeed', archive, ts, 'MAX')
+        v,t = calcMinMax('windSpeed', dbm, ts, 'MAX')
         data['windSpeed_max'] = v
         data['windSpeed_max_time'] = t
-        v,t = calcMinMax('windGust', archive, ts, 'MAX')
+        v,t = calcMinMax('windGust', dbm, ts, 'MAX')
         data['windGust_max'] = v
         data['windGust_max_time'] = t
-        v,t = calcMinMax('barometer', archive, ts, 'MAX')
+        v,t = calcMinMax('barometer', dbm, ts, 'MAX')
         data['pressure_max'] = v
         data['pressure_max_time'] = t
-        v,t = calcMinMax('barometer', archive, ts, 'MIN')
+        v,t = calcMinMax('barometer', dbm, ts, 'MIN')
         data['pressure_min'] = v
         data['pressure_min_time'] = t
-        data['10min_high_gust'] = calc10MinHighGust(archive, ts)
+        data['10min_high_gust'] = calc10MinHighGust(dbm, ts)
         ut = weewx.units.getStandardUnitType(pu, 'outTemp')
         vt = (packet.get('outTemp'), ut[0], 'group_temperature')
         v = weewx.units.convert(vt, 'degree_F')[0]
@@ -390,20 +418,23 @@ class CumulusRealTime(StdService):
         vt = (v, 'degree_C', 'group_temperature')
         v = weewx.units.convert(vt, ut[0])[0]
         data['humidex'] = v
-        data['ET_today'] = calcETToday(archive, ts)
-        data['10min_avg_wind_bearing'] = calc10MinAvgWindDir(archive, ts)
-        data['rain_hour'] = calcRainHour(archive, ts)
+        data['ET_today'] = calcETToday(dbm, ts)
+        data['10min_avg_wind_bearing'] = calc10MinAvgWindDir(dbm, ts)
+        data['rain_hour'] = calcRainHour(dbm, ts)
         data['is_daylight'] = calcIsDaylight(ts, self.latitude, self.longitude)
-        data['lost_sensors_contact'] = lostSensorContact()
-        data['avg_wind_dir'] = degreeToCompass(calcAvgWindDir(archive, ts))
+        data['lost_sensors_contact'] = lostSensorContact(packet)
+        data['avg_wind_dir'] = degreeToCompass(calcAvgWindDir(dbm, ts))
         ut = weewx.units.getStandardUnitType(pu, 'outTemp')
         vt = (packet.get('outTemp'), ut[0], 'group_temperature')
         t_F = weewx.units.convert(vt, 'degree_F')[0]
         ut = weewx.units.getStandardUnitType(pu, 'outTemp')
         vt = (packet.get('dewpoint'), ut[0], 'group_temperature')
         d_F = weewx.units.convert(vt, 'degree_F')[0]
-        data['cloud_base'] = calcCloudBase(t_F, d_F, self.altitude_ft)
-        data['units_cloud_base'] = 'ft'
+        cb_f = calcCloudBase(t_F, d_F, self.altitude_ft)
+        ut = weewx.units.getStandardUnitType(pu, 'altitude')
+        vt = (cb_f, "foot", "group_altitude")
+        data['cloud_base'] = weewx.units.convert(vt, ut[0])[0]
+        data['units_cloud_base'] = ut[0]
         if (v < 50):
             ut = weewx.units.getStandardUnitType(pu, 'outTemp')
             vt = (packet.get('outTemp'), ut[0], 'group_temperature')
